import Graphics.GD
import System.Directory
import Data.List
import Control.Concurrent
import Control.Monad
import Control.Parallel
import Control.Parallel.Strategies
import Data.Tuple

type Position = (Double, Double)


{-

This program generates Lyapunov fractal.

Works really slow.
To run, compile with:
`ghc -O2 -funfolding-use-threshold=16 -optc-O3 -optc-ffast-math -threaded --make fractal.hs`
To run:
`./fractal +RTS -N`
Wait... For over an hour on my machine...

Nice fractals to try:
iterSequence = "BBBBBBAAAAAA"
lmax = 0.7
lmin = -3

iterSequence = "ABAABABBBAA"
lmax = 1.9
lmin = -2

iterSequence = "AAB"
lmax = 2
lmin = -5.2

-}

{-
Here is the configuration:
-}

iterSequence :: String --Primary sequence. Just a string of As and Bs.
iterSequence = "AAB"

{-
Iterations to do to approximate Lyapunov exponent. The more the better.
4000 seems to be quite good, but still not perfect, any less than 600 looks horrible.
-}
iterations :: Int 
iterations = 8000

{-
Run computation in parallel?
Parallel computation is way faster on machines with more processors than sequential
However parallel computation allocates way more memory (over 5gb for 4096x4096) (Haskell's fault), so it might be a good idea to turn this off for larger images on lower memory machines
-}
runInParallel :: Bool
runInParallel = True

lmax :: Double --What is the maximum expected unstability
lmax = 2
lmin :: Double --What is the minimum expected stability
lmin = -5.2

gradss :: Color --Colour for barely stable system
gradss = rgb 248 168 70

gradse :: Color --Colour for the most stable system
gradse = rgb 102 12 36

gradus :: Color --Colour for barely unstable system
gradus = rgb 12 68 255

gradue :: Color --Colour for most stable system
gradue = rgb 7 112 87

colorsf :: Double -> Double --Colour interpolation function for stable system (maps [0..1] -> [0..1])
colorsf x = abs $ sin $ 10 * x

coloruf :: Double -> Double --Colour interpolation function for unstable system (maps [0..1] -> [0..1])
coloruf x = abs $ sin $ 10 * x


size :: Size --Size of the canvas
size = (4096, 4096)

area :: (Position, Position) --Frame to render (defined by two points in interval [0..4]x[0..4])
area = ((3.4,2.5), (4.0, 3.4))



{-
Here begins the code
-}


width :: Double
width = (fst.snd) area - (fst.fst) area

height :: Double
height = (snd.snd) area - (snd.fst) area

repeated :: String
repeated = cycle iterSequence

--This code is highly optimised. Allocating as little as possible to minimise garbage collector work.
lyexponent :: Double -> Double -> Double
lyexponent a b = (1 / (fromIntegral iterations)) * iterated
    where
        iterated = iteratedHelper iterations 0.5 repeated :: Double
            where
                iteratedHelper :: Int -> Double -> String -> Double
                iteratedHelper 0 _ _       = 0
                iteratedHelper it x (c:cs) = vf + iteratedHelper (it - 1) val cs
                    where
                        k = if c == 'A' then a else b
                        val = k * x * (1 - x) 
                        vf = f val k
                        f x v = logBase 2 $ abs (k*(1-2*val))


render :: Double -> Double -> Color
render a b = colorise $ lyexponent a b

colorise :: Double -> Color
colorise x | x < 0     = gradient colorsf gradse gradss lmin 0 x
           | otherwise = gradient coloruf gradus gradue 0 lmax x

gradient :: (Double -> Double) -> Color -> Color -> Double -> Double -> Double -> Color
gradient f c1 c2 lmin lmax val = rgb r g b
    where
        (r1, g1, b1, _) = toRGBA c1
        (r2, g2, b2, _) = toRGBA c2
        normalised = (val-lmin)/(lmax-lmin)
        interpolate :: Int -> Int -> Double -> Int
        interpolate start end val = round $ (1 - f val) * (fromIntegral start) + (f val) * (fromIntegral end)
        r = interpolate r1 r2 normalised
        g = interpolate g1 g2 normalised
        b = interpolate b1 b2 normalised

allPixels :: Int -> Int -> [Point]
allPixels w h = [(x,y) | x <- [0..(w - 1)], y <- [0..(h - 1)]]

toBounded :: Point -> Position
toBounded p = (((fst.fst) area) + width * (1 - ((fromIntegral . fst) p / (fromIntegral . fst) size)),
                ((snd.fst) area) + height * (fromIntegral . snd) p / (fromIntegral . snd) size)

--Helper function for calculating colour bounds 
minmax = do
    let coefs = map (uncurry lyexponent) $ map toBounded $ allPixels (fst size) (snd size)
    print $ minimum coefs
    print $ maximum coefs

main = do
    image <- newImage size
    let pixels = allPixels (fst size) (snd size)
    let mapFun = if runInParallel then parMap rpar else map
    let colors = mapFun (uncurry render . toBounded . swap) pixels
    mapM_ (apply image) $ zip pixels colors

    createDirectoryIfMissing True "Output/"
    savePngFile ("Output/" ++ iterSequence ++ ".png") image
    where
        apply image (pixel, color) = setPixel pixel color image
